/*
 * Port H Masks
 */

#ifndef __BFIN_PERIPHERAL_PORT_H__
#define __BFIN_PERIPHERAL_PORT_H__

#define PH0		(1 << 0)
#define PH1		(1 << 1)
#define PH2		(1 << 2)
#define PH3		(1 << 3)
#define PH4		(1 << 4)
#define PH5		(1 << 5)
#define PH6		(1 << 6)
#define PH7		(1 << 7)
#define PH8		(1 << 8)
#define PH9		(1 << 9)
#define PH10		(1 << 10)
#define PH11		(1 << 11)
#define PH12		(1 << 12)
#define PH13		(1 << 13)
#define PH14		(1 << 14)
#define PH15		(1 << 15)

#endif
