/*
 * include/asm-xtensa/user.h
 *
 * Xtensa Processor version.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 - 2005 Tensilica Inc.
 */

#ifndef _XTENSA_USER_H
#define _XTENSA_USER_H

/* This file usually defines a 'struct user' structure. However, it it only
 * used for a.out file, which are not supported on Xtensa.
 */

#endif	/* _XTENSA_USER_H */
